@@ -5,6 +5,7 @@ extern "C" {
 // https://refspecs.linuxbase.org/LSB_3.1.0/LSB-Desktop-generic/LSB-Desktop-generic/libjpegman.html
 
 #include <math.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -19,23 +20,35 @@ extern "C" {
 // #include "jmorecfg.h"
 // #endif
 
+// === FLAGS ===
 char flag_is_set (BITMASK flags, BITMASK mask) { return (flags & mask) != 0; }
 char overwrite_flag (BITMASK flags, BITMASK mask) { return (flags & (mask << 1)) == 0;}
+#define DO_FANCY_UPSAMPLING     ((BITMASK)0b1 <<  0)
+#define DO_BLOCK_SMOOTHING      ((BITMASK)0b1 <<  2)
+#define TWO_PASS_QUANTIZE       ((BITMASK)0b1 <<  4)
+#define ENABLE_1PASS_QUANT      ((BITMASK)0b1 <<  6)
+#define ENABLE_EXTERNAL_QUANT   ((BITMASK)0b1 <<  8)
+#define ENABLE_2PASS_QUANT      ((BITMASK)0b1 << 10)
+#define OPTIMIZE_CODING         ((BITMASK)0b1 << 12)
+#define PROGRESSIVE_MODE        ((BITMASK)0b1 << 14)
+#define QUANTIZE_COLORS         ((BITMASK)0b1 << 16)
+#define ARITH_CODE              ((BITMASK)0b1 << 18)
+#define WRITE_JFIF_HEADER       ((BITMASK)0b1 << 20)
+#define WRITE_ADOBE_MARKER      ((BITMASK)0b1 << 22)
+#define CCIR601_SAMPLING        ((BITMASK)0b1 << 24)
+#define FORCE_BASELINE          ((BITMASK)0b1 << 26)
 
-#define DO_FANCY_UPSAMPLING          ((BITMASK)0b1 <<  0)
-#define DO_BLOCK_SMOOTHING           ((BITMASK)0b1 <<  2)
-#define TWO_PASS_QUANTIZE            ((BITMASK)0b1 <<  4)
-#define ENABLE_1PASS_QUANT           ((BITMASK)0b1 <<  6)
-#define ENABLE_EXTERNAL_QUANT        ((BITMASK)0b1 <<  8)
-#define ENABLE_2PASS_QUANT           ((BITMASK)0b1 << 10)
-#define OPTIMIZE_CODING              ((BITMASK)0b1 << 12)
-#define PROGRESSIVE_MODE             ((BITMASK)0b1 << 14)
-#define QUANTIZE_COLORS              ((BITMASK)0b1 << 16)
-#define ARITH_CODE                   ((BITMASK)0b1 << 18)
-#define WRITE_JFIF_HEADER            ((BITMASK)0b1 << 20)
-#define WRITE_ADOBE_MARKER           ((BITMASK)0b1 << 22)
-#define CCIR601_SAMPLING             ((BITMASK)0b1 << 24)
-
+// === MARKERS ===
+// globals
+#define MAX_MARKER 20
+static int gpos = 0;
+static int gmarker_types[MAX_MARKER];
+static int gmarker_lengths[MAX_MARKER];
+static unsigned char * gmarker_data[MAX_MARKER];
+int set_marker_handlers(struct jpeg_decompress_struct * cinfo); // set up handlers
+int unset_marker_handlers(struct jpeg_decompress_struct * cinfo); // unset up handlers
+int jpeg_getc (j_decompress_ptr cinfo); // read next byte
+int jpeg_handle_marker (j_decompress_ptr cinfo); // async marker handler
 
 
 
@@ -43,13 +56,19 @@ long jround_up (long a, long b);
 
 FILE *_read_jpeg(const char *filename,
                  struct jpeg_decompress_struct *cinfo,
-                 struct jpeg_error_mgr *jerr) {
+                 struct jpeg_error_mgr *jerr,
+                 bool read_header) {
   // open file
   FILE *fp;
   if ((fp = fopen(filename, "rb")) == NULL) {
-    fprintf(stderr, "can't open %s\n", filename);
+    //fprintf(stderr, "can't open %s\n", filename);
     return NULL;
   }
+  // check file size
+  fseek(fp, 0L, SEEK_END);
+  size_t fsize = ftell(fp);
+  fseek(fp, 0L, SEEK_SET);
+  if(fsize == 0) return NULL;
 
   // zero the structures
   memset(cinfo,0x00,sizeof(struct jpeg_decompress_struct));
@@ -59,18 +78,22 @@ FILE *_read_jpeg(const char *filename,
   cinfo->err = jpeg_std_error(jerr);
   jpeg_create_decompress(cinfo);
   jpeg_stdio_src(cinfo, fp);
-  (void) jpeg_read_header(cinfo, TRUE);
+
+  if(read_header)
+    (void) jpeg_read_header(cinfo, TRUE);
 
   return fp;
 }
 
 int read_jpeg_info(
   const char *srcfile,
-  int *dct_dims,
+  int *block_dims,
   int *image_dims,
   int *num_components,
   int *samp_factor,
-  int *jpeg_color_space
+  int *jpeg_color_space,
+  int *marker_lengths,
+  int *marker_types
 ) {
   // allocate
   struct jpeg_decompress_struct cinfo;
@@ -78,28 +101,43 @@ int read_jpeg_info(
 
   // read jpeg header
   FILE *fp;
-  if((fp = _read_jpeg(srcfile, &cinfo, &jerr)) == NULL) return 0;
+  if((fp = _read_jpeg(srcfile, &cinfo, &jerr, FALSE)) == NULL) return 0;
+
+  // markers
+  if((marker_lengths != NULL) || (marker_types != NULL)) {
+    // setup
+    set_marker_handlers(&cinfo);
+    // read markers
+    (void) jpeg_read_header(&cinfo, TRUE);
+    // collect marker data
+    for(int i = 0; i < gpos; i++) {
+      marker_lengths[i] = gmarker_lengths[i];
+      marker_types[i] = gmarker_types[i];
+    }
+    // cleanup
+    unset_marker_handlers(&cinfo);
+  }
+
   jpeg_calc_output_dimensions(&cinfo);
 
   (void)jpeg_read_coefficients(&cinfo);
 
   // copy to caller
-  if(dct_dims != NULL) {
-    for(int i = 0; i < 3; i++) {
-      dct_dims[2*i] = cinfo.comp_info[i].width_in_blocks;
-      dct_dims[2*i+1] = cinfo.comp_info[i].height_in_blocks;
+  if(block_dims != NULL) {
+    for(int i = 0; i < cinfo.num_components; i++) {
+      block_dims[2*i] = cinfo.comp_info[i].height_in_blocks;
+      block_dims[2*i+1] = cinfo.comp_info[i].width_in_blocks;
     }
   }
   if(image_dims != NULL) {
-    image_dims[0] = cinfo.output_width;
-    image_dims[1] = cinfo.output_height;
+    image_dims[0] = cinfo.output_height;
+    image_dims[1] = cinfo.output_width;
   }
   if(num_components != NULL) {
     num_components[0] = cinfo.num_components;
     //num_components[1] = cinfo.out_color_components;
     //num_components[2] = cinfo.output_components;
   }
-  //fprintf(stderr, "setting jpeg color space %d %d\n", cinfo.jpeg_color_space, cinfo.out_color_space);
   if(jpeg_color_space != NULL)
     jpeg_color_space[0] = cinfo.out_color_space;
 
@@ -116,32 +154,73 @@ int read_jpeg_info(
   return 1;
 }
 
-void *_dct_offset(short * base, int channel, int w, int h, int Wmax, int Hmax)
-{
-  return (void *)(base + 64*(h + Hmax*(w + Wmax*(channel))));
-}
-
-int read_jpeg_dct(
+int read_jpeg_markers(
   const char *srcfile,
-  short *dct,
-  unsigned short *qt
-) { 
+  unsigned char *markers
+) {
   // allocate
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
 
   // read jpeg header
   FILE *fp;
-  if((fp = _read_jpeg(srcfile, &cinfo, &jerr)) == NULL) return 0;
+  if((fp = _read_jpeg(srcfile, &cinfo, &jerr, FALSE)) == NULL) return 0;
 
+  // markers
+  if(markers != NULL) {
+    // set marker handlers
+    set_marker_handlers(&cinfo);
+    // read markers
+    (void) jpeg_read_header(&cinfo, TRUE);
+    // collect marker data
+    int offset = 0;
+    for(int i = 0; i < gpos; i++) {
+      for(int j = 0; j < gmarker_lengths[i]; j++) {
+        markers[offset + j] = gmarker_data[i][j];
+      }
+      offset += gmarker_lengths[i];
+    }
+    unset_marker_handlers(&cinfo);
+  }
+
+  //(void)jpeg_read_coefficients(&cinfo);
+  // cleanup
+  jpeg_destroy_decompress( &cinfo );
+  fclose(fp);
+
+  return 1;
+}
+
+void *_dct_offset(short ** base, int channel, int h, int w, int Hmax, int Wmax)
+{
+  return (void *)(base[channel] + 64*(w + Wmax*(h + Hmax*(0))));
+}
+
+int read_jpeg_dct(
+  const char *srcfile,
+  short *Y,
+  short *Cb,
+  short *Cr,
+  unsigned short *qt
+) {
+  // allocate
+  struct jpeg_decompress_struct cinfo;
+  struct jpeg_error_mgr jerr;
+  // read jpeg header
+  FILE *fp;
+  if((fp = _read_jpeg(srcfile, &cinfo, &jerr, FALSE)) == NULL) return 0;
+  // TODO: read markers
+  (void) jpeg_read_header(&cinfo, TRUE);
   // read DCT
   jvirt_barray_ptr *coeffs_array = jpeg_read_coefficients(&cinfo);
   // read dct
   JBLOCKARRAY buffer_one;
   JCOEFPTR blockptr_one;
-  int HblocksY = cinfo.comp_info->height_in_blocks; // max height
-  int WblocksY = cinfo.comp_info->width_in_blocks; // max width
+  //int HblocksY = cinfo.comp_info->height_in_blocks; // max height
+  //int WblocksY = cinfo.comp_info->width_in_blocks; // max width
+  short *dct[3] = {Y, Cb, Cr};
   for(int ch = 0; ch < cinfo.num_components; ch++) {
+    if(dct[ch] == NULL) continue; // skip component, if null
     jpeg_component_info* compptr_one = cinfo.comp_info + ch;
     int Hblocks = compptr_one->height_in_blocks; // max height
     int Wblocks = compptr_one->width_in_blocks; // max width
@@ -149,10 +228,11 @@ int read_jpeg_dct(
       buffer_one = (cinfo.mem->access_virt_barray)((j_common_ptr)&cinfo, coeffs_array[ch], h, (JDIMENSION)1, FALSE);
       for(int w = 0; w < Wblocks; w++) {
         blockptr_one = buffer_one[0][w];
+        //fprintf(stderr, "* access: %d %d\n", ch, h, w);
         for(int bh = 0; bh < 8; bh++) {
           for(int bw = 0; bw < 8; bw++) {
             int i = bw*8 + bh;
-            ((short *)_dct_offset(dct, ch, w, h, WblocksY, HblocksY))[i] = blockptr_one[bh*8 + bw];
+            ((short *)_dct_offset(dct, ch, h, w, Hblocks, Wblocks))[i] = blockptr_one[bh*8 + bw];
           }
         }
         //memcpy(_dct_offset(dct, ch, w, h, WblocksY, HblocksY), (void *)blockptr_one, sizeof(short)*64);
@@ -161,14 +241,14 @@ int read_jpeg_dct(
   }
   // read quantization table
   if(qt != NULL) {
-    if(cinfo.num_components > 1) {
-      if(cinfo.comp_info[1].quant_tbl_no != cinfo.comp_info[2].quant_tbl_no) {
-        fprintf(stderr, "Mismatching chrominance quantization tables not supported.");
-        return 0;
-      }
-    }
+    // if(cinfo.num_components > 1) {
+    //   if(cinfo.comp_info[1].quant_tbl_no != cinfo.comp_info[2].quant_tbl_no) {
+    //     fprintf(stderr, "Mismatching chrominance quantization tables not supported.");
+    //     return 0;
+    //   }
+    // }
 
-    for(int ch = 0; ch < 2; ch++) {
+    for(int ch = 0; ch < cinfo.num_components; ch++) {
 
       //fprintf(stderr, "qt[%d] = %p -> %p\n", ch, cinfo.quant_tbl_ptrs[ch], cinfo.quant_tbl_ptrs[ch]->quantval);
       for(int i = 0; i < 64; i++) {
@@ -208,25 +288,35 @@ int read_jpeg_dct(
 int write_jpeg_dct(
   const char *srcfile,
   const char *dstfile,
-  short *dct,
+  short *Y,
+  short *Cb,
+  short *Cr,
   int *image_dims,
+  int *block_dims,
   int in_color_space,
   int in_components,
-  int *samp_factor,
   unsigned short *qt,
-  short quality
+  short quality,
+  int num_markers,
+  int *marker_types,
+  int *marker_lengths,
+  unsigned char *markers
 ) {
   // check inputs
   if(dstfile == NULL) {
-    fprintf(stderr, "destination file not specified\n");
+    fprintf(stderr, "you must specify dstfile\n");
     return 0;
   }
-  if((srcfile == NULL) && (qt == NULL) && (quality < 0)) {
-    fprintf(stderr, "you must specify either srcfile, qt or quality\n");
+  if((qt == NULL) && (quality < 0)) {
+    fprintf(stderr, "you must specify either qt or quality\n");
     return 0;
   }
-  if((srcfile == NULL) && (dct == NULL)) {
-    fprintf(stderr, "you must specify either srcfile or dct\n");
+  if(Y == NULL) {
+    fprintf(stderr, "you must specify Y\n");
+    return 0;
+  }
+  if(((Cb != NULL) && (Cr == NULL)) || ((Cb == NULL) && (Cr != NULL))) {
+    fprintf(stderr, "you must specify Y or YCbCr\n");
     return 0;
   }
 
@@ -248,18 +338,21 @@ int write_jpeg_dct(
   struct jpeg_error_mgr jerr_in;
   FILE * fp_in;
   // read source jpeg
-  if(srcfile != NULL)
-    if((fp_in = _read_jpeg(srcfile, &cinfo_in, &jerr_in)) == NULL) return 0;
+  if(srcfile != NULL) {
+    if((fp_in = _read_jpeg(srcfile, &cinfo_in, &jerr_in, FALSE)) == NULL) return 0;
+    // todo write markers
+    (void) jpeg_read_header(&cinfo_in, TRUE);
+  }
 
   cinfo_out.err = jpeg_std_error(&jerr_out);
   jpeg_create_compress(&cinfo_out);
   jpeg_stdio_dest(&cinfo_out, fp_out);
-
+  
   if(srcfile != NULL) // copy critical parameters to dstfile
     jpeg_copy_critical_parameters((j_decompress_ptr)&cinfo_in,&cinfo_out);
 
-  cinfo_out.image_width = image_dims[0];
-  cinfo_out.image_height = image_dims[1];
+  cinfo_out.image_height = image_dims[0];
+  cinfo_out.image_width = image_dims[1];
   cinfo_out.in_color_space = in_color_space;
   if(in_components >= 0) cinfo_out.input_components = in_components;
   cinfo_out.num_components = cinfo_out.input_components;
@@ -268,18 +361,18 @@ int write_jpeg_dct(
     jpeg_set_defaults(&cinfo_out);
 
   // set sampling factors
-  int chroma_factor[2];
-  if(samp_factor != NULL) {
-    chroma_factor[0] = *(samp_factor + 0);
-    chroma_factor[1] = *(samp_factor + 1);
-    for(int comp = 0; comp < cinfo_out.num_components; comp++) {
-      cinfo_out.comp_info[comp].h_samp_factor = *(samp_factor + comp*2 + 0);
-      cinfo_out.comp_info[comp].v_samp_factor = *(samp_factor + comp*2 + 1);
-    }
-  } else {
-    chroma_factor[0] = cinfo_out.comp_info[0].h_samp_factor;
-    chroma_factor[1] = cinfo_out.comp_info[0].v_samp_factor;
-  }
+  //int chroma_factor[2];
+  //if(samp_factor != NULL) {
+  //  chroma_factor[0] = *(samp_factor + 0);
+  //  chroma_factor[1] = *(samp_factor + 1);
+    //for(int comp = 0; comp < cinfo_out.num_components; comp++) {
+    //  cinfo_out.comp_info[comp].h_samp_factor = *(samp_factor + comp*2 + 0);
+    //  cinfo_out.comp_info[comp].v_samp_factor = *(samp_factor + comp*2 + 1);
+    //}
+  //} else {
+  //  chroma_factor[0] = cinfo_out.comp_info[0].h_samp_factor;
+  //  chroma_factor[1] = cinfo_out.comp_info[0].v_samp_factor;
+  //}
   // for(int comp = 0; comp < cinfo.num_components; comp++) {
   //     *(samp_factor + comp*2 + 0) = cinfo.comp_info[comp].h_samp_factor;
   //     *(samp_factor + comp*2 + 1) = cinfo.comp_info[comp].v_samp_factor;
@@ -297,11 +390,7 @@ int write_jpeg_dct(
   //   chroma_factor[0] = cinfo_out.comp_info[0].h_samp_factor;
   //   chroma_factor[1] = cinfo_out.comp_info[0].v_samp_factor;
   // }
-  //fprintf(stderr, "chroma factors %d %d\n", chroma_factor[0], chroma_factor[1]);
-  //for(int ch = 0; ch < 3; ch++) {
-  //  fprintf(stderr, "sampling factors(%d) %d %d\n", ch, cinfo_out.comp_info[ch].v_samp_factor, cinfo_out.comp_info[ch].h_samp_factor);
-  //}
-  
+
     // for(int comp = 0; comp < cinfo_out.input_components; comp++) {
     //   //cinfo_out.comp_info[comp].h_samp_factor
     //   //int J_factor,a_factor,b_factor;
@@ -345,12 +434,12 @@ int write_jpeg_dct(
       jpeg_component_info* comp_ptr = cinfo_out.comp_info + ch;
       //long v_samp_factor = *(samp_factor + ch*2 + 0);
       //long h_samp_factor = *(samp_factor + ch*2 + 1);
-      comp_ptr->width_in_blocks = (JDIMENSION)ceil(((double)cinfo_out.image_width) / 8);// * comp_ptr->h_samp_factor / 4;
-      comp_ptr->height_in_blocks = (JDIMENSION)ceil(((double)cinfo_out.image_height) / 8);// * comp_ptr->v_samp_factor / 4;
-      if(ch > 0) {
-        comp_ptr->width_in_blocks = ceil(((double)comp_ptr->width_in_blocks) / chroma_factor[0]);
-        comp_ptr->height_in_blocks = ceil(((double)comp_ptr->height_in_blocks) / chroma_factor[1]);
-      }
+      comp_ptr->height_in_blocks = (JDIMENSION)block_dims[2*ch];//(JDIMENSION)ceil(((double)cinfo_out.image_width) / 8);// * comp_ptr->h_samp_factor / 4;
+      comp_ptr->width_in_blocks = (JDIMENSION)block_dims[2*ch+1];//ceil(((double)cinfo_out.image_height) / 8);// * comp_ptr->v_samp_factor / 4;
+      //if(ch > 0) {
+      //  comp_ptr->width_in_blocks = ceil(((double)comp_ptr->width_in_blocks) / chroma_factor[0]);
+      //  comp_ptr->height_in_blocks = ceil(((double)comp_ptr->height_in_blocks) / chroma_factor[1]);
+      //}
       coeffs_array[ch] = (cinfo_out.mem->request_virt_barray)(
         (j_common_ptr)&cinfo_out,
         JPOOL_IMAGE,
@@ -368,31 +457,40 @@ int write_jpeg_dct(
   #if JPEG_LIB_VERSION >= 80
   jpeg_calc_jpeg_dimensions(&cinfo_out);
   #endif
-  //fprintf(stderr, "before writing coefficients\n");
+
   jpeg_write_coefficients(&cinfo_out,coeffs_array);
-  //fprintf(stderr, "written coefficients\n");
+
+  // write markers
+  int offset = 0;
+  for(int i = 0; i < num_markers; i++) {
+    //fprintf(stderr, "- %d type %d length %d offset %d data [%d %d %d %d ...]\n", i, marker_types[i], marker_lengths[i], offset, markers[offset], markers[offset+1], markers[offset+2], markers[offset+3]);
+    jpeg_write_marker(&cinfo_out, marker_types[i], markers + offset, marker_lengths[i]);
+    offset += marker_lengths[i];
+  }
+
   // write DCT coefficients
-  if(dct != NULL) {
-    JBLOCKARRAY buffer_one;
-    JCOEFPTR blockptr_one;
-    int HblocksY = cinfo_out.comp_info->height_in_blocks; // max height
-    int WblocksY = cinfo_out.comp_info->width_in_blocks; // max width
-    for(int ch = 0; ch < 3; ch++) { // channel iterator
-      jpeg_component_info* comp_ptr = cinfo_out.comp_info + ch;
-      int Hblocks = comp_ptr->height_in_blocks; // max height
-      int Wblocks = comp_ptr->width_in_blocks; // max width
-      for(int h = 0; h < Hblocks; h++) { // height iterator
-        //fprintf(stderr, "accessing ch%d h%d/[H%d,W%d]\n", ch, h, Hblocks, Wblocks);
-        buffer_one = (cinfo_out.mem->access_virt_barray)((j_common_ptr)&cinfo_out, coeffs_array[ch], h, (JDIMENSION)1, TRUE);
-        for(int w = 0; w < Wblocks; w++) { // width iterator
-          blockptr_one = buffer_one[0][w];
-          for(int bh = 0; bh < 8; bh++)
-            for(int bw = 0; bw < 8; bw++)
-              blockptr_one[bh*8 + bw] = ((short *)_dct_offset(dct, ch, w, h, WblocksY, HblocksY))[bw*8 + bh];
-        }
+  JBLOCKARRAY buffer_one;
+  JCOEFPTR blockptr_one;
+  //int HblocksY = cinfo_out.comp_info->height_in_blocks; // max height
+  //int WblocksY = cinfo_out.comp_info->width_in_blocks; // max width
+  short *dct[3] = {Y, Cb, Cr};
+  for(int ch = 0; ch < 3; ch++) { // channel iterator
+    if(dct[ch] == NULL) continue;
+    jpeg_component_info* comp_ptr = cinfo_out.comp_info + ch;
+    int Hblocks = comp_ptr->height_in_blocks; // max height
+    int Wblocks = comp_ptr->width_in_blocks; // max width
+    for(int h = 0; h < Hblocks; h++) { // height iterator
+      //fprintf(stderr, "accessing ch%d h%d/[H%d,W%d]\n", ch, h, Hblocks, Wblocks);
+      buffer_one = (cinfo_out.mem->access_virt_barray)((j_common_ptr)&cinfo_out, coeffs_array[ch], h, (JDIMENSION)1, TRUE);
+      for(int w = 0; w < Wblocks; w++) { // width iterator
+        blockptr_one = buffer_one[0][w];
+        for(int bh = 0; bh < 8; bh++)
+          for(int bw = 0; bw < 8; bw++)
+            blockptr_one[bh*8 + bw] = ((short *)_dct_offset(dct, ch, h, w, Hblocks, Wblocks))[bw*8 + bh];
       }
     }
   }
+  
   // cleanup
   jpeg_finish_compress( &cinfo_out );
   jpeg_destroy_compress( &cinfo_out );
@@ -423,18 +521,20 @@ int read_jpeg_spatial(
 
   // read jpeg header
   FILE *fp;
-  if((fp = _read_jpeg(srcfile, &cinfo, &jerr)) == NULL) return 0;
+  if((fp = _read_jpeg(srcfile, &cinfo, &jerr, TRUE)) == NULL) return 0;
 
-  // set parameters
+  // set parameterse
   if(out_color_space >= 0) cinfo.out_color_space = out_color_space;
   else cinfo.out_color_space = cinfo.jpeg_color_space;
-  //fprintf(stderr, "cjpeglib.c: out_color_space: %d %d\n", out_color_space, cinfo.out_color_space);
-  
   if(dither_mode >= 0) cinfo.dither_mode = dither_mode;
   if(dct_method >= 0) cinfo.dct_method = dct_method;
 
-  if (overwrite_flag(flags, DO_FANCY_UPSAMPLING))
+  if (overwrite_flag(flags, DO_FANCY_UPSAMPLING)) {
     cinfo.do_fancy_upsampling = flag_is_set(flags, DO_FANCY_UPSAMPLING);
+    #if JPEG_LIB_VERSION >= 70
+    fprintf(stderr, "- read %s: DO_FANCY_UPSAMPLING %d min_DCT_h_scaled_size %d max_DCT_h_scaled_size %d\n", srcfile, cinfo.do_fancy_upsampling, cinfo.min_DCT_h_scaled_size, cinfo.min_DCT_h_scaled_size);
+    #endif
+  }
   if (overwrite_flag(flags, DO_BLOCK_SMOOTHING))
     cinfo.do_block_smoothing  = flag_is_set(flags, DO_BLOCK_SMOOTHING);
   if (overwrite_flag(flags, QUANTIZE_COLORS))
@@ -444,8 +544,6 @@ int read_jpeg_spatial(
   if(in_colormap != NULL)
     for(int i = 0; i < 256; i++) {
       cmap[i] = in_colormap + i*3;
-      //if(i < 3) fprintf(stderr, " %d %d |", i, cmap[i][0]);
-      //if(i == 255) fprintf(stderr, "\n");
     }
 
   if (overwrite_flag(flags, QUANTIZE_COLORS) && flag_is_set(flags, QUANTIZE_COLORS)) {
@@ -453,29 +551,7 @@ int read_jpeg_spatial(
     cinfo.desired_number_of_colors = 256;
     if(in_colormap != NULL) cinfo.colormap = (char**)cmap;
   }
-  
-  //fprintf(stderr, "read PROGRESSIVE_MODE: owrt %d set %d\n", 
-  //  overwrite_flag(flags, PROGRESSIVE_MODE),
-  //  flag_is_set(flags, PROGRESSIVE_MODE)
-  //);fprintf(stderr, "ARITH_CODE: owrt %d set %d\n", 
-  //   overwrite_flag(flags, ARITH_CODE),
-  //   flag_is_set(flags, ARITH_CODE)
-  // );fprintf(stderr, "CCIR601_SAMPLING: owrt %d set %d\n", 
-  //   overwrite_flag(flags, CCIR601_SAMPLING),
-  //   flag_is_set(flags, CCIR601_SAMPLING)
-  // );fprintf(stderr, "TWO_PASS_QUANTIZE: owrt %d set %d\n", 
-  //   overwrite_flag(flags, TWO_PASS_QUANTIZE),
-  //   flag_is_set(flags, TWO_PASS_QUANTIZE)
-  // );fprintf(stderr, "ENABLE_1PASS_QUANT: owrt %d set %d\n", 
-  //   overwrite_flag(flags, ENABLE_1PASS_QUANT),
-  //   flag_is_set(flags, ENABLE_1PASS_QUANT)
-  // );fprintf(stderr, "ENABLE_EXTERNAL_QUANT: owrt %d set %d\n", 
-  //   overwrite_flag(flags, ENABLE_EXTERNAL_QUANT),
-  //   flag_is_set(flags, ENABLE_EXTERNAL_QUANT)
-  // );fprintf(stderr, "ENABLE_2PASS_QUANT: owrt %d set %d\n", 
-  //   overwrite_flag(flags, ENABLE_2PASS_QUANT),
-  //   flag_is_set(flags, ENABLE_2PASS_QUANT)
-  // );
+
   if (overwrite_flag(flags, PROGRESSIVE_MODE))
     cinfo.progressive_mode      = flag_is_set(flags, PROGRESSIVE_MODE);
   if (overwrite_flag(flags, ARITH_CODE))
@@ -494,7 +570,7 @@ int read_jpeg_spatial(
   // decompress
   (void)jpeg_start_decompress(&cinfo);
   // read pixels
-  unsigned char *rowptr = rgb;
+  char *rowptr = (char *)rgb;
   unsigned short stride = (overwrite_flag(flags, QUANTIZE_COLORS) && flag_is_set(flags, QUANTIZE_COLORS)) ?
     1 : cinfo.out_color_components;
   
@@ -522,7 +598,6 @@ int read_jpeg_spatial(
 }
 
 int write_jpeg_spatial(
-  const char *srcfile,
   const char *dstfile,
   unsigned char *rgb,
   int *image_dims,
@@ -533,6 +608,10 @@ int write_jpeg_spatial(
   unsigned short *qt,
   short quality,
   short smoothing_factor,
+  int num_markers,
+  int *marker_types,
+  int *marker_lengths,
+  unsigned char *markers,
   BITMASK flags
 ) {
 
@@ -549,19 +628,16 @@ int write_jpeg_spatial(
   cinfo.err = jpeg_std_error(&jerr);
   jpeg_create_compress(&cinfo);
   jpeg_stdio_dest(&cinfo, fp);
-
+  
   // set basic parameters
-  if(srcfile == NULL) {
-    cinfo.image_width = image_dims[0];
-    cinfo.image_height = image_dims[1];
-    if(in_color_space >= 0)
-      cinfo.in_color_space = in_color_space;
-    //fprintf(stderr, "writing 1 from cs %d to %d\n", cinfo.in_color_space, cinfo.jpeg_color_space);
-    if(in_components >= 0)
-      cinfo.input_components = in_components;
-    cinfo.num_components = cinfo.input_components;
-    jpeg_set_defaults(&cinfo);
-  }
+  cinfo.image_height = image_dims[0];
+  cinfo.image_width = image_dims[1];
+  if(in_color_space >= 0)
+    cinfo.in_color_space = in_color_space;
+  if(in_components >= 0)
+    cinfo.input_components = in_components;
+  cinfo.num_components = cinfo.input_components;
+  jpeg_set_defaults(&cinfo);
 
   // set advanced parameters
   if(dct_method >= 0) cinfo.dct_method = dct_method;
@@ -578,19 +654,6 @@ int write_jpeg_spatial(
     chroma_factor[1] = cinfo.comp_info[0].v_samp_factor;
   }
 
-  // copy parameters
-  struct jpeg_decompress_struct cinfo_in;
-  struct jpeg_error_mgr jerr_in;
-  FILE * fp_in = NULL;
-  if(srcfile != NULL) {
-    // read file
-    if((fp_in = _read_jpeg(srcfile, &cinfo_in, &jerr_in)) == NULL) return 0;
-    // decompress
-    (void)jpeg_start_decompress(&cinfo_in);
-    jpeg_copy_critical_parameters((j_decompress_ptr)&cinfo_in, (j_compress_ptr)&cinfo);
-  }
-
-  //fprintf(stderr, "quality %d\n", quality);
   if(qt != NULL) {
     unsigned qt_u[64];
     // component 0
@@ -607,8 +670,14 @@ int write_jpeg_spatial(
     cinfo.comp_info[2].component_id = 2;
     cinfo.comp_info[2].quant_tbl_no = 1;
   } else if(quality >= 0) {
-    jpeg_set_quality(&cinfo, quality, FALSE);
+    // force baseline (8bit quantization)
+    bool force_baseline = FALSE;
+    if (overwrite_flag(flags, FORCE_BASELINE))
+      force_baseline = flag_is_set(flags, FORCE_BASELINE);
+    // set quality
+    jpeg_set_quality(&cinfo, quality, force_baseline);
   }
+
   if(smoothing_factor >= 0) cinfo.smoothing_factor = smoothing_factor;
   if(in_color_space >= 0)
     cinfo.in_color_space = in_color_space;
@@ -617,15 +686,16 @@ int write_jpeg_spatial(
   if (overwrite_flag(flags, DO_FANCY_UPSAMPLING))
     cinfo.do_fancy_downsampling = flag_is_set(flags, DO_FANCY_UPSAMPLING);
   #endif
-  //fprintf(stderr, "write PROGRESSIVE_MODE owr %d set %d\n", overwrite_flag(flags, PROGRESSIVE_MODE), flag_is_set(flags, OPTIMIZE_CODING));
   if (overwrite_flag(flags, PROGRESSIVE_MODE))
     cinfo.progressive_mode   = flag_is_set(flags, PROGRESSIVE_MODE);
   if (overwrite_flag(flags, PROGRESSIVE_MODE) && flag_is_set(flags, PROGRESSIVE_MODE))
     jpeg_simple_progression(&cinfo);
   if (overwrite_flag(flags, OPTIMIZE_CODING))
     cinfo.optimize_coding    = flag_is_set(flags, OPTIMIZE_CODING);
+  #ifdef C_ARITH_CODING_SUPPORTED
   if (overwrite_flag(flags, ARITH_CODE))
     cinfo.arith_code         = flag_is_set(flags, ARITH_CODE);
+  #endif
   if (overwrite_flag(flags, WRITE_JFIF_HEADER))
     cinfo.write_JFIF_header  = flag_is_set(flags, WRITE_JFIF_HEADER);
   if (overwrite_flag(flags, WRITE_ADOBE_MARKER))
@@ -633,24 +703,26 @@ int write_jpeg_spatial(
   if (overwrite_flag(flags, CCIR601_SAMPLING))
     cinfo.CCIR601_sampling   = flag_is_set(flags, CCIR601_SAMPLING);
 
-  // https://gist.github.com/kentakuramochi/f64e7646f1db8335c80f131be8359044
+  // start compression
+  jpeg_start_compress(&cinfo, TRUE);
+
+  // write markers
+  int offset = 0;
+  for(int i = 0; i < num_markers; i++) {
+    jpeg_write_marker(&cinfo, marker_types[i], markers + offset, marker_lengths[i]);
+    offset += marker_lengths[i];
+  }
 
   // write data
-  unsigned char *rowptr = rgb;
-  jpeg_start_compress(&cinfo, TRUE);
+  char *rowptr = (char *)rgb;
   for(unsigned h = 0; h < cinfo.image_height; h++) {
     jpeg_write_scanlines(&cinfo, &rowptr, 1);
     rowptr += cinfo.image_width * cinfo.input_components;
   }
-
   // cleanup
   jpeg_finish_compress( &cinfo );
   jpeg_destroy_compress( &cinfo );
   fclose( fp );
-  if(srcfile != NULL) {
-    jpeg_destroy_decompress( &cinfo_in );
-    fclose( fp_in );
-  }
   
   return 1;
 }
@@ -664,7 +736,7 @@ int print_jpeg_params(const char *srcfile)
 
   // read jpeg header
   FILE *fp;
-  if((fp = _read_jpeg(srcfile, &cinfo, &jerr)) == NULL) return 0;
+  if((fp = _read_jpeg(srcfile, &cinfo, &jerr, TRUE)) == NULL) return 0;
 
   (void)jpeg_start_decompress(&cinfo);
 
@@ -725,6 +797,105 @@ int print_jpeg_params(const char *srcfile)
   return 1;
 }
 
+int set_marker_handlers(struct jpeg_decompress_struct * cinfo) {
+  // jpeg globals
+  gpos = 0;
+  for(int i = 0; i < MAX_MARKER; i++) {
+    gmarker_types[i] = 0;
+    gmarker_data[i] = NULL;
+    gmarker_lengths[i] = 0;
+  }
+  // set handlers
+  jpeg_set_marker_processor(cinfo, JPEG_COM, jpeg_handle_marker);
+  jpeg_set_marker_processor(cinfo, JPEG_APP0+15, jpeg_handle_marker);
+  for (int i=1; i<14; i++)
+    jpeg_set_marker_processor(cinfo, JPEG_APP0+i, jpeg_handle_marker);
+  
+  return 1;
+}
+
+int unset_marker_handlers(struct jpeg_decompress_struct * cinfo) {
+  // jpeg globals
+  for(int i = 0; i < gpos; i++) {
+    gmarker_lengths[i] = 0;
+    gmarker_types[i] = 0;
+    if(gmarker_data[i] != NULL)
+      free((void*)gmarker_data[i]);
+  }
+  gpos = 0;
+  
+  // set handlers
+  (void)cinfo;
+
+  return 1;
+}
+
+/**
+ * Read next byte.
+ */
+int jpeg_getc (j_decompress_ptr cinfo)
+{
+  struct jpeg_source_mgr * datasrc = cinfo->src;
+  // no bytes in buffer
+  if (datasrc->bytes_in_buffer == 0) {
+    if (! (*datasrc->fill_input_buffer) (cinfo))
+      return -1; // return error
+  }
+  // read char
+  datasrc->bytes_in_buffer--;
+  return GETJOCTET(*datasrc->next_input_byte++);
+}
+
+/**
+ * Asynchronous marker handler.
+ */
+int jpeg_handle_marker (j_decompress_ptr cinfo)
+{
+  // get marker name
+  char mname[20];
+  if (cinfo->unread_marker == JPEG_COM) sprintf(mname, "COM");
+  else sprintf(mname, "APP%d", cinfo->unread_marker - JPEG_APP0);
+  
+  // get length
+  unsigned char * p = NULL;
+  int length = 0;
+  length = jpeg_getc(cinfo) << 8;
+  length += jpeg_getc(cinfo);
+  length -= 2; // discount the length word itself
+  gmarker_lengths[gpos] = length;
+
+  // allocate
+  if(gpos < MAX_MARKER) {
+    gmarker_types[gpos] = cinfo->unread_marker;
+    //strcpy(mark_name[gpos], mname);
+    if((p = malloc((length + 1)*sizeof(char))) == NULL) {
+      fprintf(stderr, "Bad malloc!\n");
+      return FALSE;
+    }
+    gmarker_data[gpos] = p;
+    gpos += 1;
+
+  // too many markers
+  } else {
+    fprintf(stderr, "Too many markers - %s skipped\n", mname);
+  }
+
+  // iterate over data
+  int c;
+  while (--length >= 0) {
+    if((c = jpeg_getc(cinfo)) == -1) {
+      fprintf(stderr, "Error parsing marker %s\n", mname);
+      return FALSE;
+    }
+    *(p) = (unsigned char)c;
+    //if (!p[0]) p[0] = 0x20; // replace 0x0 byte with 0x20
+    p++;
+  }
+  p[0] = 0; // set the last byte to 0
+  
+  return TRUE;
+}
+
 
 
 #ifdef __cplusplus
